"""
index.py

Indexing module for Yet Another FASTA Index and Extraction Tool
"""

import sys
import argparse
from pathlib import Path
from yafax.core import buildIndex, writeIndex

def runIndex(args: argparse.Namespace) -> None:
    """
    Runs the index module
    """
    fasta = args.fasta_path
    outdir = args.out_dir
    if not outdir:
        outdir = None
    if not fasta.exists():
        print(f"FASTA file {fasta} does not exist!", file = sys.stderr)
        sys.exit(2)
    index = buildIndex(fasta)
    writeIndex(index, fasta, outdir)

if __name__ == "__main__":
    runIndex()



"""
cli.py

CLI interface for Yet Another FASTA Index and Extraction Tool

Parses args and exposes the CLI interface
"""

import argparse
from pathlib import Path
from yafax import __version__
from yafax.index import runIndex
from yafax.getseq import runGetseq

def buildParser() -> argparse.ArgumentParser:
    """
    Function to setup argparse and fetch args.
    """
    parser = argparse.ArgumentParser(
            prog = "yafax",
            description = "YAFAX (Yet Another FASTA Indexing and Extraction Tool)",
            usage = "\tyafax <mode> --help\n\tyafax <mode> [mode optional arguments] <fasta file>",
            allow_abbrev = False
    )

    parser.add_argument(
            "-v",
            "--version",
            action = "version",
            version = f"%(prog)s {__version__}"
    )

    # Create subparsers for modules
    mode_parser = parser.add_subparsers(
            title = "yafax mode",
            dest = "mode",
            required = True
    )

    INDEX_parser = mode_parser.add_parser(
            "index",
            help = "generate a FASTA index",
            usage = "yafax index [optional arguments] <fasta file>"
    )

    # Attach function
    INDEX_parser.set_defaults(func = runIndex)

    GETSEQ_parser = mode_parser.add_parser(
            "getseq",
            help = "fetch sequence of an interval using position",
            usage = "yafax getseq [optional arguments] <position> <fasta file>"
    )

    # Attach function
    GETSEQ_parser.set_defaults(func = runGetseq)

    parser.add_argument(
            "fasta_path",
            type = Path,
            metavar = "FASTA FILE",
            help = "path to FASTA file"
    )

    INDEX_parser.add_argument(
            "-d",
            "--out-dir",
            type = Path,
            required = False,
            metavar = "",
            help = "directory of the output FASTA index"
    )

    GETSEQ_parser.add_argument(
            "-i",
            "--index-path",
            type = Path,
            required = False,
            metavar = "",
            help = "path of FASTA index (Default: file.fa.fai)"
    )

    GETSEQ_parser.add_argument(
            "-r",
            "--revcomp",
            default = False,
            action = "store_true",
            help= "reverse complement the sequence (Default: False)"
    )

    GETSEQ_parser.add_argument(
            "-w",
            "--width",
            type = int,
            default = 50,
            metavar = "",
            help = "wrapping width of FASTA lines (Default: 50)"
    )

    GETSEQ_parser.add_argument(
            "-n",
            "--no-header",
            default = False,
            action = "store_true",
            help = "do not include FASTA header (Default: False)"
    )

    GETSEQ_parser.add_argument(
            "-b",
            "--bedfile",
            type = Path,
            metavar = "",
            required = False,
            help = "BED file with regions"
    )

    GETSEQ_parser.add_argument(
            "-o",
            "--outfile",
            type = Path,
            metavar = "",
            required = False,
            help = "output file, only used when BED is provided (Default: STDOUT)"
    )

    GETSEQ_parser.add_argument(
            "position",
            type = str,
            nargs = "?",
            help = "position interval in format chr:start-end"
    )

    return parser

def cli() -> None:
    """
    CLI entrypoint
    """
    parser = buildParser()
    args = parser.parse_args()
    args.func(args)

if __name__ == "__main__":
    cli()



"""
test_getseq.py

Test script to run test for YAFAX getseq
"""

import sys
from pathlib import Path
import pytest
import subprocess

def test_getseq() -> bool:
    """
    Uses chrm.fa as the genome and chrm.fa.fai
    as the index. Both are stored in tests/data

    Runs yafax getseq and checks whether it matches
    with tests/data/chrm_10000_11000.fa
    """
    position = "chrM:10000-11000"
    genome = Path("tests/data/chrm.fa")
    expected_result = Path("tests/data/chrm_10000_11000.fa")

    cli_cmd = ["yafax", "getseq", position, str(genome)] 
    out = subprocess.run(cli_cmd, capture_output = True, text = True, check = True)
    assert out.stdout == expected_result.read_text()

if __name__ == "__main__":
    pytest.main()



"""
test_index.py

Script to test yafax index module
"""

import sys
import subprocess
import pytest
from pathlib import Path

def test_index(tmp_path: Path) -> bool:
    """
    Uses tests/data/chrm.fa as input for
    indexing. If the indexing happens it
    will check whether it matches the expected
    result
    """
    genome = Path("tests/data/chrm.fa")
    expected_index = Path("tests/data/chrm.fa.fai")
    faidir = tmp_path / "results"
    faidir.mkdir()

    cli_cmd = ["yafax", "index", "--out-dir", str(faidir), str(genome)]
    subprocess.run(cli_cmd, capture_output = True, check = True, text = True)
    
    produced_index = faidir / "chrm.fa.fai"
    
    assert produced_index.exists()
    assert produced_index.read_text() == expected_index.read_text()

if __name__ == "__main__":
    pytest.main()



"""
logger.py

Logging module for faidex

Can be used by calling:
    from logger import logMessage as logger

Example:
    logger("warning", f"File {file }is missing")
"""
import sys
import logging

# Configure the logger
_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)

# Create console handler
_console_handler = logging.StreamHandler(stream = sys.stderr)
_console_handler.setLevel(logging.DEBUG)

# Set format
_formatter = logging.Formatter(fmt = "[%(asctime)s] %(levelname)s %(message)s", datefmt = "%Y-%m-%d %H:%M:%S")

# Add formatting to console handler
_console_handler.setFormatter(_formatter)

# Add console handler to logger
if not _logger.handlers:
    _logger.addHandler(_console_handler)

def logMessage(level: str, message: str) -> None:
    """
    Portable function to expose logging to the
    main script. Required level and message can
    be passed to it for debug. Always writes to
    STDERR.

    Levels: debug, info, warning, error, critical
    """
    level = level.lower()
    log_func = getattr(_logger, level, None)

    if callable(log_func):
        log_func(message)
    else:
        _logger.warning("Unknown level '%s', message: %s", level, message)



"""
getseq.py

getseq module for Yet Another FASTA Index and Extraction Tool
"""

import sys
import argparse
from pathlib import Path
from yafax.core import revComp, readBed, readIndex, writeSeq, fetchSeq, fetchSeqBed

def runGetseq(args: argparse.Namespace) -> None:
    """
    Runs the getseq module
    """
    width = args.width
    isrevcomp = args.revcomp
    isheader = args.no_header
    fasta = args.fasta_path

    # Check FASTA path
    if not fasta.exists():
        print(f"FASTA file {fasta} does not exist!", file = sys.stderr)
        sys.exit(2)

    # Check index path
    index = args.index_path
    if not index:
        index = fasta.with_suffix(fasta.suffix + ".fai")
    if not index.exists():
        print(f"Index {index} does not exist!, try indexing first", file = sys.stderr)
        sys.exit(2)

    bedfile = args.bedfile
    position = args.position
    outfile = args.outfile

    # Check if --bedfile and <position> supplied together
    if bedfile and position:
        print("Cannot use <position> when --bedfile is provided", file = sys.stderr)
        sys.exit(1)
    elif not bedfile and not position:
        print("Must provide either <position> or --bedfile", file = sys.stderr)
        sys.exit(1)

    # if --bedfile is provided run fetchSeqBed else if <position> is valid run fetchSeq
    if bedfile:
        if not bedfile.exists():
            print(f"BED file {bedfile} does not exist!", file = sys.stderr)
            sys.exit(2)
        if not outfile:
            print("Outfile not provided, writing to STDOUT", file = sys.stderr)
            outfile = None
        index_dict = readIndex(index)
        fetchSeqBed(fasta, index_dict, bedfile, width, isrevcomp, isheader, outfile)
    else:
        try:
            chrom, ival = position.strip().split(":")
            start_str, end_str = ival.split("-")
            start = int(start_str)
            end = int(end_str)
        except ValueError:
            print("Position must be in the format chrom:start-end", file = sys.stderr)
            sys.exit(1)
        if start < 0 or end < 0 or start >= end:
            print("Start and end must be positive and start < end", file = sys.stderr)
            sys.exit(1)
        index_dict = readIndex(index)
        fetchSeq(fasta, index_dict, chrom, int(start), int(end), width, isrevcomp, isheader)

if __name__ == "__main__":
    runGetseq()



"""
core.py

Core functions for Yet Another FASTA Index and Extraction Tool
"""

import sys
from pathlib import Path
from typing import Iterator, Optional
from yafax.logger import logMessage as logger

def revComp(seq: str) -> str:
    """
    Reverse complements a given sequence
    """
    complement = str.maketrans("ACTGNactgn", "TGACNtgacn")
    return seq.translate(complement)[::-1]

def readBed(bed_path: Path) -> Iterator[tuple[str, int, int]]:
    """
    Reads a BED file and generates positions as a tuple
    """
    with open(bed_path, mode = "r", encoding = "utf-8") as bh:
        for line in bh:
            if line.startswith("#") or line.strip() == "":
                continue
            chrom, start, end, *_  = line.rstrip().split("\t")
            yield chrom, int(start), int(end)

def buildIndex(fasta_path: Path) -> dict[str, tuple[int, int | None, int | None, int | None]]:
    """
    Takes a fasta file as input and returns a dictionary
    of index fields.
    """
    index = {}
    with fasta_path.open(mode = "rb") as fh:
        seq_name = None
        seq_len = 0
        seq_offset = None
        bases_per_line = None
        bytes_per_line = None

        while True:
            line_offset = fh.tell()
            line = fh.readline()
            if not line:
                break
            if line.startswith(b">"):
                if seq_name is not None:
                    logger("info", f"Creating index for contig: {seq_name}")
                    index[seq_name] = (seq_len, seq_offset, bases_per_line, bytes_per_line, )
                seq_name = line[1:].strip().split()[0].decode()
                seq_len = 0
                seq_offset = None
                bases_per_line = None
                bytes_per_line = None
                continue
            if seq_offset is None:
                seq_offset = line_offset
            clean = line.rstrip(b"\n")
            line_clean = len(clean)
            if bases_per_line is None:
                bases_per_line = line_clean
                bytes_per_line = len(line)
            seq_len += line_clean
        if seq_name is not None:
            index[seq_name] = (seq_len, seq_offset, bases_per_line, bytes_per_line)
    return index

def writeIndex(index: dict[str, tuple[int, int | None, int | None, int | None]], fasta_path: Path, outdir: Path | None) -> None:
    """
    Takes dictionary of index fields and writes to
    a tab-delimited index file.
    """
    if outdir is not None:
        outdir.mkdir(parents = True, exist_ok = True)
        index_path = outdir / (fasta_path.name + ".fai")
    else:
        index_path = fasta_path.with_suffix(fasta_path.suffix + ".fai")
    if index_path.exists():
        raise FileExistsError(f"Index already exists: {index_path}")
    with index_path.open(mode = "w", encoding = "utf-8") as out:
        for name, (length, offset, bpl, bypl) in index.items():
            out.write(f"{name}\t{length}\t{offset}\t{bpl}\t{bypl}\n")

def readIndex(index_path: Path) -> dict[str, tuple[int, int, int, int]]:
    """
    Reads the FASTA index and returns a tuple of
    chrom and byte fields.
    """
    index = {}
    with index_path.open(mode = "r", encoding = "utf-8") as fh:
        for line in fh:
            chrom, length, offset, bpl, bypl = line.rstrip().split("\t")
            index[chrom] = (int(length), int(offset), int(bpl), int(bypl))
    return index

def fetchSeq(fasta_path: Path, index: dict[str, tuple[int, int, int, int]], chrom: str, start: int, end: int, width: int, revcomp: bool, noheader: bool) -> None:
    """
    Fetches sequence of an interval using the fasta
    index and fasta sequence itself (position is 1-based).
    """
    length, offset, bpl, bypl = index[chrom]
    if end > length:
        raise ValueError(f"Requested end {end} > chromosome length {length}!")
    # Convert to 0-based positions
    start0 = start - 1
    start_line = start0 // bpl
    start_col = start0 % bpl
    bytes_pos = offset + start_line * bypl + start_col
    to_read = end - start0
    lines = (start_col + to_read + bpl - 1) // bpl
    bytes_to_read = lines * bypl
    with fasta_path.open(mode = "rb") as fh:
        fh.seek(bytes_pos)
        seq_bin = fh.read(bytes_to_read)
    seq = seq_bin.replace(b"\n", b"")[:to_read].decode()
    header: str | None = None if noheader else f">{chrom}:{start}-{end}\n"
    writeSeq(seq = seq, header = header, width = width, revcomp = revcomp)

def fetchSeqBed(fasta_path: Path, index: dict[str, tuple[int, int, int, int]], bed_path: Path, width: int = 50, revcomp: bool = False, noheader: bool = False, outfile: Optional[Path] = None) -> None:
    """
    Fetches sequences from intervals of BED entries
    and writes to outfile or STDOUT
    """
    out = sys.stdout if outfile is None else outfile.open(mode = "w", encoding = "utf-8")
    # BED path validity will be provided by the upper level function, assumes correct BED
    try:
        with fasta_path.open(mode = "rb") as fh:
            for chrom, start, end in readBed(bed_path):
                if start < 1 or end < 1 or end < start:
                    logger("warning", f"Invalid BED interval {chrom}:{start}-{end}")
                    continue
                length, offset, bpl, bypl = index[chrom]
                if end > length:
                    logger("warning", f"Requested end {end} > chromosome length {length}!")
                    continue
                logger("info", f"Fetching sequence for {chrom}:{start}-{end}")
                start0 = start - 1
                start_line = start0 // bpl
                start_col = start0 % bpl
                bytes_pos = offset + start_line * bypl + start_col
                to_read = end - start0
                lines = (start_col + to_read + bpl - 1) // bpl
                bytes_to_read = lines * bypl
                fh.seek(bytes_pos)
                seq_bin = fh.read(bytes_to_read)
                seq = seq_bin.replace(b"\n", b"")[:to_read].decode()
                if revcomp:
                    seq = revComp(seq)
                out_lines: list[str] = []
                if not noheader:
                    out_lines.append(f">{chrom}:{start}-{end}\n")
                for pos in range(0, len(seq), width):
                    out_lines.append(seq[pos: pos + width] + "\n")
                out.write("".join(out_lines))
    finally:
        if outfile is not None:
            out.close()

def writeSeq(seq: str, header: None | str, width: int = 50, revcomp: bool = False) -> None:
    """
    Figure out yourself what it does!
    """
    out = sys.stdout
    if revcomp:
        seq = revComp(seq)
    if header is not None:
        out.write(header)
    for pos in range(0, len(seq), width):
        out.write(seq[pos:pos + width] + "\n")
